
package engine;

import memory.*;

import java.io.*;
import java.lang.reflect.*;

/**
Klasse der Virtuellen Maschine V2M.
Copyright (C) 2006, Rene Heuer, Salingtwiete 4g. (D)20535 Hamburg
eMail: heuer@exmpl.de
Dieses Programm ist freie Software. Sie können es unter den Bedingungen der
GNU General Public License, wie von der Free Software Foundation veröffentlicht,
weitergeben und/oder modifizieren, entweder gemäß Version 2 der Lizenz oder
(nach Ihrer Option) jeder späteren Version.
Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, daß es Ihnen von
Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite
Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK.
Details finden Sie in der GNU General Public License.
Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem
Programm erhalten haben. Falls nicht, schreiben Sie an die
Free Software Foundation, Inc.,
59 Temple Place,
Suite 330,
Boston, MA 02111-1307,
USA.
*/
public class V2M {
/**
Innstruction-Pointer. Ist das Analogon zu einem "Program-Counter" in Prozessoren.
*/
public Item IP  = null;

/**
Daten-Stack. Zentrale Datenspeicher bei Stackmaschinen.
*/
public List DS  = new List();

/**
Return-Stack. Gleiche Funktion, wie der Stack bei "normalen" Prozessoren.
*/
public List RS  = new List();

/**
Dictionary. Wie Programmbereich im RAM bei "normalen" Prozessoren. Hier sind alle
compilierten Commands, Variablen und Methoden gespeichert.
*/
public List DICT = new List();

/**
Definitionsbereich. Neue Compilierungen finden hier statt.
War die Compilation erfolgreich, gelangt der belegte Bereich ins Dictionary.
*/
public List HERE = new List();

/**
Step counter to find the jump target in a command sequence
*/
public int count = 0;

/**
Aktueller Reader der Eingabe. Er kann an eine Datei oder an einen String gebunden
werden. Normalerweise ist es der String deraktuellen Eingabe.
*/
public Reader inp;

/**
Name des Eingabestroms. Normalerweise der Name einer Datei.
*/
public String inpnam;

/**
Stack zur Koordinierung verschachtelter Dateien bei der Eingabe.
*/
public List loadstack = new List();

 /**
 Klasse zur Bereitstellung einer Struktur um den Zusammenhang von Reader und Name des
 aktuellen Eingabestroms zu speichern. Instanzen dieser Klasse liegen auf dem Stack
 zur Koordinierung erschachtelter Dateieingaben.
 */
 private class lcb {
  Reader reader;
  String name;
 }

 /**
 Test, ob Eingabedatei bereits in Benutzung.
 @param name Name der Eingabedatei.
 @return true wenn in Berabeitung, sonst false.
 */
 public boolean inuse( String name) {
  if (name.equals( inpnam)) return true;

 Item item = loadstack.first;
  while (item != null)
        {
        String exst = ((lcb)item.content).name; // Name einer benutzten Datei
         if (exst != null)
            if (name.equals( exst)) return true;
         item = item.next;
        }
  return false;
 }

 /**
 Sichern des aktuellen Eingabestroms.
 */
 public void remind() {
 lcb block = new lcb();

  block.reader = inp;
  block.name   = inpnam;
  List.push( loadstack, block);
 }

 /**
 Ladenn des gemerkten Eingabestroms.
 */
 public void remember() {
 lcb block = (lcb)List.pop( loadstack);

  inp    = block.reader;
  inpnam = block.name;
 }

 /**
 Reader bereitstellen.
 @param source Ist entweder String oder Datei.
 @return Einen Reader der entsprechenden Quelle.
 */
 public Reader reader( Object source) {
  if (source instanceof String)
     {
      return new StringReader( (String)source); // StringReader
     }
  if (!(source instanceof File))
     {
      System.out.println( "Quelle nicht definiert.");
      return null;
     }
  try    {
          return new FileReader( (File)source); // FileReader
          }
  catch (IOException ior)
         {
          System.out.println( "Zugriff auf "+((File)source).getName()+"verweigert.");
          return null;
         }
 }

 /**
 Datei zum Lesen bereitstellen.
 @param name Name der Datei.
 @return File-Objekt oder null, wenn nicht vorhanden oder nicht zu lesen.
 */
 public File file( String name) {
 File fil = new File( name);

  if (!fil.exists())
     {
      System.out.println( "Datei "+name+" nicht gefunden.");
      return null;
     }
  if (!fil.canRead())
     {
      System.out.println( "Datei "+name+" kann nicht gelesen werden.");
      return null;
     }
  return fil;
 }

 /**
 Nicht-leere Zeichenfolge aus Eingabestrom isolieren.
 @return Isolierte Zeichenfolge.
 */
 public String read() {
 String str = "";

 try    {
         int chr;

          if (((chr = inp.read()) >= 0)) // Zeichen vorhanden?
             {
              if (Character.isWhitespace( (char)chr)) // skip white spaces?
                  while (((chr = inp.read()) >= 0) && (Character.isWhitespace( (char)chr)));
              str = str.concat( new String( new char[] {(char)chr}));
              // String?
              if (str.startsWith( "\"")) // String?
                 {
//                  while (((chr = inp.read()) >= 0) && ((char)chr != '"'))
//                          str = str.concat( new String( new char[] {(char)chr}));
                  while (((chr = inp.read()) >= 0) && ((char)chr != '"'))
                        {
                         if ((char)chr == '\\')
                            {
                            int chr1 = inp.read();
                             if (chr1 < 0) return str.concat( new String( new char[] {(char)chr}));
                             switch ((char)chr1) {
                              case 'b':  chr = (int)'\b'; break;
                              case 'n':  chr = (int)'\n'; break;
                              case 'r':  chr = (int)'\r'; break;
                              case 't':  chr = (int)'\t'; break;
                              case '\"': chr = (int)'\"'; break;
                              case '\'': chr = (int)'\''; break;
                             }
                            }
                         str = str.concat( new String( new char[] {(char)chr}));
                        }
                  return str;
                 }
              // Character?
              if (str.startsWith( "'")) // Leerzeichen als Character ?
              {
               while (((chr = inp.read()) >= 0) && ((char)chr != '\''))
                   str = str.concat( new String( new char[] {(char)chr}));
               if ((char)chr == '\'') str = str.concat( new String( new char[] {(char)chr}));
               return str;
              }

              while (((chr = inp.read()) >= 0) && (!Character.isWhitespace( (char)chr))) // word
                      str = str.concat( new String( new char[] {(char)chr}));
              if (str.startsWith( "//")) // Komentar?
                 {
                  if ((char)chr == ' ') str = str.concat( new String( new char[] {(char)chr})); // ' ' zum Kommentar
                  while (((chr = inp.read()) >= 0) && ((char)chr != '\n'))
                          str = str.concat( new String( new char[] {(char)chr}));
                 }
              // Anachronismen werden hier abgefangen
              if ((str.length() == 1) && (str.charAt( 0) == 0xFFFF)) str = null;
              return str;
             }
         }
  catch (IOException ior)
         {
          System.out.println( "Fehler bei Lesevorgang.\nVerarbeitung wird abgebrochen.");
          System.exit( 1);
         }
  return null;
 }

 /**
 Nicht-leere Zeichenfolge isolieren.
 Es wird auch der durch "load" unterbrochene Strom benutzt.
 @return Isolierte Zeichenfolge.
 */
 public String getWord() {
 String str = read();

  if ((str == null) && (loadstack.first != null)) // Nichts mehr da?
     {
      remember();   // Erinnern an übergeordnete Eingabe?
      str = read(); // weiterlesen
     }
  return str;
 }

 /**
 Laden einer Datei.
 Weil Dateien von der V2M sowohl zum Compilieren als auch zum Interpretieren
 oder Beides verstanden werden, muss rekursive Benutzung ausgeschlossen werden.
 Rekursion kann hier auch indirekt auftreten, weshalb ein erheblicher Aufwand
 zur Vermeidung dieser nicht abzubrechenden Situation betrieben wird.
 @param name Name der Datei.
 */
 public void runLOAD( String name) {
  if (inuse( name))
     {
      System.out.println( "Rekursive Nutzung von Dateien scheitert an nicht existierendem Abbruchkriterium.");
      System.out.println( "System wird verlassen.");
      System.exit( 1);
     }
  remind();
  inp    = reader( file( name));
  inpnam = name;
 }

 /**
 Command starten.
 Der InstructionPointer (IP) muss auf einem Listenelement liegen,
 dessen Inhalt (content) ein Command der V2M darstellt.
 */
 public void execute() {
  while (IP != null)
         {
          List.push( RS, IP.next);              // InstructionPointer auf folgendes Command und auf ReturnStack ablegen
          ((Command)IP.content).runTime( this); // Laufzeitverhalten des Commands aktivieren
          IP = (Item)List.pop( RS);             // InstructionPointer vom ReturnStack holen
         }
 }

 /**
 Anweisung im Dictionary suchen.
 @param word Bezeichnung der gesuchten Anweisung.
 @return Gefundene Anweisung oder null, wenn nicht gefunden.
 */
 public Command getCommand( String word) {
 Item    item = DICT.first;
 Command cmd  = null;

  while (item != null)
        {
         cmd = (Command)item.content;
         if (word.equals( cmd.name)) return cmd;
         item = item.next;
        }
  return null;
 }

 /**
 Word in Symbol-Command (oft bei Zahlen) wandeln.
 @param word Zu symbolisierende Zeichenfolge.
 @return Command "_sym" oder null, wenn keine Symbolisierung gefunden.
 */
 public Command getSymbol( String word) {
  if (word.equals( "null")) return new _sym( "null_sym", null);
  try { return new _sym( new Byte( word)); } // Byte-Instanz [-(2**7-1) ... 2**7]
  catch ( NumberFormatException no_byte)
          {
           try { return new _sym( new Short( word)); } // Short-Instanz [-(2**15-1) ... 2**15]
           catch ( NumberFormatException no_short)
                   {
                    try { return new _sym( new Integer( word)); } // Integer-Instanz [-(2**31-1) ... 2**31]
                    catch ( NumberFormatException no_integer)
                           {
                            try { return new _sym( new Long( word)); } // Long-Instanz [-(2**63-1) ... 2**63]
                            catch ( NumberFormatException no_long)
                                   {
                                    try { return new _sym( new Float( word)); } // Float-Instanz [-(2**127) ... 2**127]
                                    catch ( NumberFormatException no_float)
                                           {
                                            try { return new _sym( new Double( word)); } // Double-Instanz [-(2**1023) ... 2**1023]
                                            catch ( NumberFormatException no_double)
                                                   {
                                                    if (word.equals( "false") || word.equals( "true"))
                                                       return new _sym( new Boolean( word)); // Boolean Instanz
                                                    if (word.startsWith( "\""))
                                                       return new _sym( new String( word.substring( 1))); // String-Instanz
                                                    if (word.startsWith( "'"))
                                                       {
                                                        if (!word.endsWith( "'")) return null;
                                                        word = word.substring( 1, word.length()-1); // Character-Instanz
                                                        if (word.length() == 1)
                                                           return new _sym( new Character( word.charAt( 0)));
                                                        if (!word.startsWith( "\\u")) return null;
                                                        return new _sym( new Character( (char)Integer.parseInt( word.substring( 2), 16)));
                                                       }
                                                   }
                                           }
                                   }
                           }
                   }
          }
  return null;
 }

 /**
 Starten eines in der Eingabe vorliegenden Programms.
 */
 public void process() {
 String  word;
 Command cmd;

  while (true) {
   word = getWord();                  // Word isolieren
   if( word == null) return;          // wenn keine weiteren words -> fertig
   if (!word.startsWith( "//"))       // Kommentar? (Text bereits von getWord() verarbeitet)
      {
       cmd = getCommand( word);       // Command?
       if   (cmd != null)
             cmd.runTime( this);      // starten
       else {
             cmd = getSymbol( word);  // validierbar?
             if  (cmd != null)
                  cmd.runTime( this); // starten
             else System.out.println( "\nWas ist das denn?  [" + word + "]\n");
            }
      }
  }
 }

 /**
 Objekt "lesbar" ausgeben.
 Es handelt sich prinzipiell um die Methode toString, jedoch um
 die speziellen Eigenheiten der V2M erweitert.
 @param object "lesbar" auszugebenden Objekt.
 @return "lesbarer" String.
 */
 public String readable( Object object) {
  if (object == null) return "null";
  if (!(object instanceof Command))
     {
      if ((object instanceof Boolean))   return object.toString();
	  if ((object instanceof Number))    return object.toString();
	  if ((object instanceof Character)) return object.toString();
      if ((object instanceof String))    return object.toString();
     String str = object.toString();
      if (str.startsWith( "class")) return "("+str+")";
 	  str = object.getClass().getName();
      return str;
     }
  try   {
        String typ = ((Command)object).name;
         // flow control
        if (typ.equals( "[jumpf]")) return typ+"("+((_jump)object).index +")";
        if (typ.equals( "[jumpt]")) return typ+"("+((_jump)object).index +")";
        if (typ.equals(  "[jump]")) return typ+"("+((_jump)object).index +")";
        if (typ.equals(  "[back]")) return typ+"("+((_back)object).index +")";
         // symbols
         if (object instanceof _sym)
            {
            String ret = ((Command)object).commands.first.content.toString();
             if (typ.equals( "[n_sym]")) return typ+"("+     ret     +")";
             if (typ.equals( "[b_sym]")) return typ+"("+     ret     +")";
             if (typ.equals( "[s_sym]")) return typ+"("+"\""+ret+"\""+")";
             if (typ.equals( "[c_sym]")) return typ+"("+"'" +ret+"'" +")";
                                         return typ+"("+     ret     +")";
            }
         return typ;
        }
  catch (ClassCastException e1)
        {
         return object.getClass().getName();
        }
 }

 /**
 Aufbau einer kompilierten Anweisung zeigen.
 @param command Anzuzeigende Anweisung.
 */
 public void showDef( Command command) {
  System.out.println( ": "+command.name);
  if (command.commands == null)
     {
      System.out.println( "  Native Code of subengine.\n");
//      System.out.println( "  Native Code der untergeordneten Maschine.\n");
      return;
     }
 Item   item = command.commands.first;
 int    step = 0;
 String stp;
  while (item != null)
         {
          stp = " "+(step++)+":";
          for ( int j=0; j<4-stp.length(); j++) System.out.print( " ");
          System.out.print( stp);
          System.out.println( " "+readable( (Command)item.content));
          item = item.next;
         }
  System.out.println( ";");
 }

 /**
 Typus (Klasse) eines Datenobjekts aus Bezeichnung ermitteln.
 @param name Bezeichnung des Datenobjekts.
 @return Klasse des Objekts.
 */
 public Class getType( String name) {
 Class cl  = null;
 int   dim = 0;

  // Dimensionen eines angenommenen Array-Types ermitteln
  while (name.endsWith( "[]"))
        {
         dim++;
         name = name.substring( 0, name.lastIndexOf( "[]"));
        }

  // Test, ob primitve Type
  if (name.equals( "boolean")) return boolean.class;
  if (name.equals( "byte"))    return byte.class;
  if (name.equals( "short"))   return short.class;
  if (name.equals( "int"))     return int.class;
  if (name.equals( "long"))    return long.class;
  if (name.equals( "float"))   return float.class;
  if (name.equals( "double"))  return double.class;
  if (name.equals( "char"))    return char.class;

  // Test auf klassifizierten Type
  try   {
         cl = Class.forName( name);
        }
  catch (ClassNotFoundException cnf)
        {
         if (name.indexOf( '.') < 0) name = "java.lang.".concat( name);
        }

  // "try again" mit java.lang-Vokabular
  try   {
         cl = Class.forName( name);
        }
  catch (ClassNotFoundException cnf)
        {
         System.out.println( "Klasse ["+name+"] nicht gefunden");
        }

  // Test auf Array-Type
  if (cl != null && dim > 0) cl = Array.newInstance( cl, new int[dim]).getClass();
  return cl;
 }

 /**
 Bestimmtes Field (Klassenvariable) einer bestimmten Klasse ermitteln.
 @param ocl Klasse mit gesuchtem Field.
 @param name Name des gesuchten Fields.
 @return Gefundenes Field oder null, wenn nicht gefungden.
 */
 public Field field( Class ocl, String name) {
  try   {
         return ocl.getField( name);
        }
  catch (NoSuchFieldException nsf)
        {
         System.out.println( "Field ["+name+"] nicht gefunden");
        }
  catch (SecurityException    se)
        {
         System.out.println( "Field ["+name+"] unterliegt Security Management");
        }
  return null;
 }

 /**
 Typecasting eines Objekts auf dem Datenstack.
 @param to Bezeichnung des Typus (z.B. int, double).
 */
 public void cast( String to) {
 Object object = List.pop( DS);
 Method method = null;

  try   {
         method = object.getClass().getMethod( to+"Value", null);
         List.push( DS, method.invoke( object, null));
        }
  catch (NoSuchMethodException nsm)
        {
         System.out.println( "Wandlung fehlgeschlagen.");
        }
  catch (IllegalAccessException    acc)
        {
         System.out.println( "Unerlaubter Zugriff auf Wandlungsmethode.");
        }
  catch (IllegalArgumentException  iae)
        {
         System.out.println( "Argumente der Wandlung fehlerhaft");
        }
  catch (InvocationTargetException ite)
        {
         System.out.println( "Wandlung erzeugt Exception");
        }
 }

 /**
 Array mit Klassen { wichtig bei argument list} erstellen.
 @param args String mit Klassenbezeichnungen (Types) durch Leerzeichen getrennt.
 @return Array mit Klassen.
 */
 public Class[] argClasses( String args) {
  if (args.length()<1) return null;
 java.util.StringTokenizer st      = new java.util.StringTokenizer( args);
 Class[]                   classes = new Class[ st.countTokens()]; // Class[] der Argumentliste
 int                       idx     = 0;

  while (st.hasMoreTokens())
        {
         classes[idx++] = getType( st.nextToken());
        }
  return classes;
 }

 /**
 Instanz einer Klasse mit default-Creator erstellen.
 */
 public void instance( Class ocl) {
  try   {
         DS.first.content = ocl.newInstance();
        }
  catch (InstantiationException ie)
        {
         System.out.println( "Klasse ["+ocl.getName()+"] konnte nicht instanziert werden");
        }
  catch (IllegalAccessException iae)
        {
         System.out.println( "Klasse ["+ocl.getName()+"] ist nicht instanzierbar");
        }
 }

 /**
 Instanz einer Klasse mit Argumentliste erstellen.
 */
 public void construct( Constructor cstr) {
 int      len  = cstr.getParameterTypes().length; // Argumentanzahl ermitteln
 Object[] data = new Object[len];                 // Array der Argumente instanzieren

  for (int i=0; i<len; i++)                       // Argumentliste vom Stack holen
       data[len-1 - i] = List.pop( DS);

  try   {
         List.push( DS, cstr.newInstance( data));
        }
  catch (InstantiationException    ie)
        {
         System.out.println( "Instanzierung verweigert");
        }
  catch (IllegalAccessException    iac)
        {
         System.out.println( "Auf Methode zur Instanzierung konnte nicht zugegriffen werden");
        }
  catch (IllegalArgumentException  iae)
        {
         System.out.println( "Argumente der Instanzierung ungeeignet");
        }
  catch (InvocationTargetException ite)
        {
         System.out.println( "Instanzierung verweigert");
        }
 }

 /**
 Bestimmte Methode starten.
 */
 public void invokation(  Object inst, Method meth) {
 int      len  = meth.getParameterTypes().length; // Argumentanzahl ermitteln
 Object[] data = new Object[len];                 // Array der Argumenliste instanzieren

  for (int i=0; i<len; i++)                       // Argumentliste vom Stack holen
       data[len-1 - i] = List.pop( DS);

  try   {
         if    (Modifier.isStatic( meth.getModifiers())) // Test, ob statische Methode
               {
               if  (meth.getReturnType() == void.class) // Test, ob returning "void"
                    meth.invoke( null, data);
               else List.push( DS, meth.invoke( null, data));
               }
         else {
               if  (meth.getReturnType() == void.class) // Test, ob returning "void"
                    meth.invoke( inst, data);
               else List.push( DS, meth.invoke( inst, data));
              }
        }
  catch (IllegalAccessException    acc)
        {
         System.out.println( "Unerlaubter Zugriff auf"+meth.getName());
        }
  catch (IllegalArgumentException  iae)
        {
         System.out.println( "Argumente von "+meth.getName()+" fehlerhaft");
        }
  catch (InvocationTargetException ite)
        {
         System.out.println( "Methode "+meth.getName()+" erzeugt Exception");
        }
 }

 /**
 Creator der V2M.
 Der elementare Sprachschatz wird im Dictionary bereitgestellt.
 */
 public V2M() {
  // self reference
  List.push( DICT, new _environment()); // environment
	
  // object and command handling
  List.push( DICT, new _symbol());   // symbol
  List.push( DICT, new _find());     // find
  List.push( DICT, new _execute());  // execute

  // data stack manipulation
  List.push( DICT, new _drop());     // drop
  List.push( DICT, new _dup());      // dup
  List.push( DICT, new _swap());     // swap
  List.push( DICT, new _ndrop());    // ndrop
  List.push( DICT, new _ndup());     // ndup
  List.push( DICT, new _nswap());    // nswap

  // creation and compiling
  List.push( DICT, new _entry());    // entry
  List.push( DICT, new _compile());  // compile
  List.push( DICT, new _bind());     // bind
  List.push( DICT, new _forget());   // forget
  List.push( DICT, new _variable()); // variable
  List.push( DICT, new _is());       // is
  List.push( DICT, new _free());     // free

  // flow control
  List.push( DICT, new _if());       // if
  List.push( DICT, new _else());     // else
  List.push( DICT, new _endif());    // endif
  List.push( DICT, new _loop());     // loop
  List.push( DICT, new _leave());    // leave
  List.push( DICT, new _endloop());  // endloop
  List.push( DICT, new _exit());     // exit

  // boolean
  List.push( DICT, new _not());      // not
  List.push( DICT, new _or());       // or
  List.push( DICT, new _and());      // and

  // relations
  List.push( DICT, new _null());     // ?null
  List.push( DICT, new _eq());       // =
  List.push( DICT, new _le());       // <=
  List.push( DICT, new _gt());       // >
  List.push( DICT, new _ge());       // >=
  List.push( DICT, new _lt());       // <
  List.push( DICT, new _ne());       // <>

  // arithmetics
  List.push( DICT, new _add());      // +
  List.push( DICT, new _sub());      // -
  List.push( DICT, new _mul());      // *
  List.push( DICT, new _div());      // -

  // in- and output
  List.push( DICT, new _word());     // word
  List.push( DICT, new _load());     // load
  List.push( DICT, new _print());    // .
  List.push( DICT, new _newline());  // newline
  List.push( DICT, new _typ());      // typ
  List.push( DICT, new _commands()); // commands
  List.push( DICT, new _stack());    // .S
  List.push( DICT, new _shwdef());   // showDef

  // SubEngine Interface (Java-Interface)

  // SubEngine Objects
  List.push( DICT, new _classify());    // classify
  List.push( DICT, new _instantiate()); // instantiate
  List.push( DICT, new _creator());     // creator
  List.push( DICT, new _build());       // build
  List.push( DICT, new _store());       // store
  List.push( DICT, new _fetch());       // fetch
  List.push( DICT, new _call());        // call
  List.push( DICT, new _method());      // method
  List.push( DICT, new _field());       // field
  List.push( DICT, new _set());         // set

  // Type casting
  List.push( DICT, new _byte());        // (byte)
  List.push( DICT, new _short());       // (short)
  List.push( DICT, new _int());         // (int)
  List.push( DICT, new _long());        // (long)
  List.push( DICT, new _float());       // (float)
  List.push( DICT, new _double());      // (double)
 }
} // end of class V2M
