
package engine;

import memory.*;

/**
Klasse aller Symbolen (valide Objekte) der V2M.
Copyright (C) 2006, Rene Heuer, Salingtwiete 4g. (D)20535 Hamburg
eMail: heuer@exmpl.de
Dieses Programm ist freie Software. Sie können es unter den Bedingungen der
GNU General Public License, wie von der Free Software Foundation veröffentlicht,
weitergeben und/oder modifizieren, entweder gemäß Version 2 der Lizenz oder
(nach Ihrer Option) jeder späteren Version.
Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, daß es Ihnen von
Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite
Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK.
Details finden Sie in der GNU General Public License.
Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem
Programm erhalten haben. Falls nicht, schreiben Sie an die
Free Software Foundation, Inc.,
59 Temple Place,
Suite 330,
Boston, MA 02111-1307,
USA.
*/
public class _sym extends Command {

 /**
 Laufzeitverhalten des Commands.
 Die V2M arbeitet streng objektiorientiert, weshalb es keine "Werte" an sich gibt.
 Zahlen, Strings und Wahrheiten (true/false) müssen als Objekte vorliegen. Um
 dem Anwender die Arbeit zu erleichtern, werden derartige Eingaben vom
 Compiler in die entsprechenden Objekte (Symbole) automatisch gewandelt.
 Symbol wurde als Begriff benutzt, weil die Zeichenfolge 123 symbolisch für
 den Wert hundertdreiundzwanzig steht und '\u00F6' symbolisch für den
 Buchstaben Ä.
 @param machine Verbindung zur V2M.
 */
 public void runTime( V2M machine) {
  List.push( machine.DS, this.commands.first.content);
 }

 /**
 Creator mit Namensvergabe.
 Hier sind spezifische Symbolisierungen erlaubt.
 @param name Bezeichnung des Sybols.
 @param obj Zu symbolisierendes Objekt.
 */
 public _sym( String name, Object obj) {
  this.commands = new List();
  List.append( this.commands, obj);
  this.name = name;
 }

 /**
 Automatischer Creator.
 Es wird zwischen numerisch, booleschen und textuellen Symbolen unterschieden.
 @param obj Zu symbolisierendes Objekt.
 */
 public _sym( Object obj) {
  this.commands = new List();
       if (obj instanceof Number)    this.name = "[n_sym]"; // numeric
  else if (obj instanceof Boolean)   this.name = "[b_sym]"; // boolean
  else if (obj instanceof String)    this.name = "[s_sym]"; // string
  else if (obj instanceof Character) this.name = "[c_sym]"; // character
  List.append( this.commands, obj);
 }
}
